﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Threading;
using System.Threading.Tasks;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using Magify.Features;
using Magify.Model;
using Magify.Types;
using UniRx;
using UnityEngine;
using UnityEngine.Networking;
using EventType = Magify.Model.EventType;

namespace Magify
{
    internal class MagifyClient : ICancelable
    {
        internal const string CountersFolderName = "counters";
        internal const string GeneralPrefsName = "general.bin";
        internal const string OffersPrefsName = "lto.bin";

        private static readonly MagifyLogger _logger = MagifyLogger.Get();

        private readonly string _configPath;
        private readonly CompositeDisposable _disposables = new();
        private readonly SubsystemsCollection _subsystems = new();

        private readonly PlatformAPI _platform;
        private readonly ContextFileStorage _contextFileStorage;

        private readonly SessionCounter _sessionCounter;
        private readonly GeneralPrefs _generalPrefs;
        private readonly ServerApi _serverApi;
        private readonly PurchaseTracker _purchaseTracker;
        private readonly AnalyticsTracker _analyticsTracker;
        private readonly AnalyticsSessionTracker _analyticsSessionTracker;
        private readonly AnalyticsService _analyticsService;
        private readonly CampaignsTracker _campaignsTracker;
        private readonly CampaignsProvider _campaignProvider;
        private readonly CampaignsCollection _campaignsCollection;
        private readonly LimitsHolder _limitsHolder;
        private readonly FeaturesProvider _featuresProvider;
        private readonly DefaultProductsContainer _defaultProducts;
        private readonly ContentProvider _contentProvider;
        private readonly AppInstallsTracker _appInstallsTracker;
        private readonly CampaignUpdateHandler _campaignUpdateHandler;
        private readonly LtoCampaignManager _ltoCampaignManager;

        private bool _isSetup;

        public bool IsDisposed => _disposables.IsDisposed;
        public IFeatures Features => _featuresProvider;
        internal IReactiveProperty<bool> IsGeoIpEnabled { get; }
        internal IReactiveProperty<Environment> Environment { get; }

        internal IReactiveProperty<SubscriptionStatus> SubscriptionStatus => _generalPrefs.SubscriptionStatus;
        internal IReactiveProperty<InAppStatus> InAppStatus => _generalPrefs.InAppStatus;
        internal IReactiveProperty<AuthorizationStatus> AuthorizationStatus => _generalPrefs.AuthorizationStatus;
        internal IReactiveProperty<string> AdjustId => _generalPrefs.AdjustId;
        internal IReactiveProperty<string> FirebaseInstanceId => _generalPrefs.FirebaseInstanceId;
        internal IReactiveProperty<string> ReferrerId => _generalPrefs.ReferrerId;
        internal IReactiveProperty<bool> IsGdprGranted => _generalPrefs.IsGdprGranted;
        internal IReadOnlyReactiveProperty<string> GpsAdId => _generalPrefs.GpsAdId;
        internal IReadOnlyReactiveProperty<bool?> AttAuthorizationStatus => _generalPrefs.AttAuthorizationStatus;

        internal bool IsTablet => _platform.IsTablet;

        internal bool IsPortrait => Screen.width < Screen.height;

        internal ContextSyncTime LastContextSyncTime { get; private set; }

        internal IReadOnlyReactiveProperty<TimeSpan> SessionsInterval => _sessionCounter.Interval;

        internal string ClientId => _generalPrefs.ClientId;
        internal LimitsModel Limits => _limitsHolder.GetActiveLimits();

        internal IReadOnlyReactiveProperty<int> SessionNumber => _generalPrefs.GlobalSessionCounter;
        internal string FirstInstalledVersion => _generalPrefs.FirstInstalledVersion;

        internal List<InAppProduct> InAppProducts => _defaultProducts.InAppProducts;

        internal List<SubscriptionProduct> SubscriptionProducts => _defaultProducts.SubscriptionProducts;

        internal PlatformAPI Platform => _platform;

        internal event Action OnConfigLoaded;
        internal event Action OnConfigSyncRequested;

        public event Action<LtoInfo> OnOfferAdded
        {
            add => _ltoCampaignManager.OnAdded += value;
            remove => _ltoCampaignManager.OnAdded -= value;
        }

        public event Action<LtoInfo> OnOfferUpdated
        {
            add => _ltoCampaignManager.OnUpdated += value;
            remove => _ltoCampaignManager.OnUpdated -= value;
        }

        public event Action<LtoInfo> OnOfferRemoved
        {
            add => _ltoCampaignManager.OnRemoved += value;
            remove => _ltoCampaignManager.OnRemoved -= value;
        }

        public event Action<LtoInfo> OnOfferFinished
        {
            add => _ltoCampaignManager.OnFinished += value;
            remove => _ltoCampaignManager.OnFinished -= value;
        }

        internal event Action OnPurchasedProductsChanged;

        // ToDo: native screenshot tracking implementation
        internal event Action OnUserDidTakeScreenshot;
        internal event Action OnApplicationEnterForeground;
        internal event Action OnApplicationEnterBackground;

        public MagifyClient(string appName, string configPath, string rootPath, Environment environment, EditorDevice editorDevice = null)
        {
            _configPath = configPath;
            Environment = new ReactiveProperty<Environment>(environment);

            _subsystems.AddTo(_disposables);
            _platform = PreparePlatformBridge(editorDevice, rootPath).AddTo(_subsystems);
            _contextFileStorage = new ContextFileStorage(rootPath).AddTo(_subsystems);
            _generalPrefs = GeneralPrefs.Create(Path.Combine(rootPath, GeneralPrefsName), _platform).AddTo(_subsystems);

            var offersPrefs = OffersPrefs.Create(Path.Combine(rootPath, OffersPrefsName)).AddTo(_subsystems);
            var countersStorage = new CountersStorage(Path.Combine(rootPath, CountersFolderName)).AddTo(_subsystems);
            var counters = new Counters(countersStorage, AlternativeCountersGetter).AddTo(_subsystems);
            var appVersion = _platform.ValidateAppVersion(Application.version);
            var authBuilder = new AuthorizationConfigBuilder(appName, appVersion, _platform, _generalPrefs).AddTo(_subsystems);

            _campaignsCollection = new CampaignsCollection().AddTo(_subsystems);

            _serverApi = new ServerApi(GetEnvironmentUrl(Environment.Value), new NetworkClient(), _generalPrefs, authBuilder, _platform).AddTo(_subsystems);
            _analyticsTracker = new AnalyticsTracker(_serverApi, _generalPrefs, rootPath).AddTo(_subsystems);
            _analyticsSessionTracker = new AnalyticsSessionTracker(_analyticsTracker, _generalPrefs, _platform).AddTo(_subsystems);

            _sessionCounter = new SessionCounter(_generalPrefs, _platform).AddTo(_subsystems);
            _limitsHolder = new LimitsHolder(_generalPrefs, _sessionCounter).AddTo(_subsystems);
            _campaignsTracker = new CampaignsTracker(_generalPrefs, counters, _limitsHolder, _platform).AddTo(_subsystems);
            _ltoCampaignManager = new LtoCampaignManager(_campaignsTracker, _campaignsCollection, _generalPrefs, offersPrefs, _platform, counters, rootPath).AddTo(_subsystems);
            _featuresProvider = new FeaturesProvider(_generalPrefs).AddTo(_subsystems);
            _campaignProvider = new CampaignsProvider(_generalPrefs, _campaignsTracker, _campaignsCollection, counters, _limitsHolder, _platform, _ltoCampaignManager).AddTo(_subsystems);

            _purchaseTracker = new PurchaseTracker(_serverApi, _platform, rootPath).AddTo(_subsystems);
            _analyticsService = new AnalyticsService(_analyticsTracker, _campaignsTracker).AddTo(_subsystems);

            _defaultProducts = new DefaultProductsContainer().AddTo(_subsystems);
            _contentProvider = new ContentProvider().AddTo(_subsystems);
            _appInstallsTracker = new AppInstallsTracker(_analyticsTracker, _generalPrefs, _platform).AddTo(_subsystems);
            _campaignUpdateHandler = new CampaignUpdateHandler(_campaignsTracker, _campaignProvider, _campaignsCollection, _platform, _generalPrefs, _ltoCampaignManager).AddTo(_subsystems);

            IsGeoIpEnabled = _serverApi.IsGeoIpEnabled;
        }

        [CanBeNull]
        private int? AlternativeCountersGetter(CounterType type, CounterScope scope, CounterKey key)
        {
            if (!_generalPrefs.MigratedUser) return null;
            try
            {
                return _platform.GetGlobalCounterFromLegacyData(type, key);
            }
            catch (Exception e)
            {
                _logger.LogException(e);
            }
            return null;
        }

        public void InitializeSdk()
        {
            try
            {
                if (_generalPrefs.ClientId == null)
                {
                    var legacyData = _platform.LoadLegacyNativeDataForMigration();
#if MAGIFY_VERBOSE_LOGGING
                    try
                    {
                        // Remove after tests
                        if (!Directory.Exists(_contextFileStorage.RootFolderPath))
                            Directory.CreateDirectory(_contextFileStorage.RootFolderPath);
                        File.WriteAllText(Path.Combine(_contextFileStorage.RootFolderPath, "migrationdata.json"), legacyData?.ToString() ?? "Empty");
                    }
                    catch (Exception e)
                    {
                        _logger.LogException(e);
                    }
#endif
                    if (legacyData != null)
                    {
                        _subsystems.ForEach<ILegacyMigrator>(c => c.Migrate(legacyData));
                        // _platform.CleanLegacyNativeDataAfterMigration();
                    }
                }
            }
            catch (Exception e)
            {
                _logger.LogException(new MagifyMigrationFailedException("All", string.Empty, e));
            }

            try
            {
                SetupSavedContext();
            }
            catch (Exception e) { _logger.LogException(new MagifyFailedToLoadContextFromDiskException(ContextKind.Saved, e)); }
            try
            {
                SetupDefaultContext(_configPath);
            }
            catch (Exception e) { _logger.LogException(new MagifyFailedToLoadContextFromDiskException(ContextKind.Default, e)); }

            if (_generalPrefs.VersionSessionCounter == 0)
            {
                PrepareForFirstSessionInVersion();
            }

            Environment
                .SkipLatestValueOnSubscribe()
                .Subscribe(value => _serverApi.ChangeServerUrl(GetEnvironmentUrl(value)))
                .AddTo(_disposables);
            _platform.OnLocaleChanged
                .Subscribe(localeData => _serverApi.RefreshLocale(localeData).IfTrue(_serverApi.CancelAllContextLoadings, Sync))
                .AddTo(_disposables);
            _platform.OnForeground
                .Subscribe(_ => OnApplicationEnterForeground?.Invoke())
                .AddTo(_disposables);
            _platform.OnBackground
                .Subscribe(_ => OnApplicationEnterBackground?.Invoke())
                .AddTo(_disposables);
            _generalPrefs.AttAuthorizationStatus
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => CancelRequestsAndSync().Forget())
                .AddTo(_disposables);
            _generalPrefs.SessionId
                .SkipLatestValueOnSubscribe()
                .Subscribe(_ => Sync())
                .AddTo(_disposables);
#if MAGIFY_VERBOSE_LOGGING
            _platform.OnForeground
                .Subscribe(_ => _logger.Log($"{nameof(_platform.OnForeground)} called"))
                .AddTo(_disposables);
            _platform.OnBackground
                .Subscribe(_ => _logger.Log($"{nameof(_platform.OnBackground)} called"))
                .AddTo(_disposables);
            _platform.OnApplicationDidBecomeActive
                ?.Subscribe(_ => _logger.Log($"{nameof(_platform.OnApplicationDidBecomeActive)} called"))
                .AddTo(_disposables);
            _platform.OnApplicationWillResignActive
                ?.Subscribe(_ => _logger.Log($"{nameof(_platform.OnApplicationWillResignActive)} called"))
                .AddTo(_disposables);
#endif

            _campaignsTracker.OnPurchasedProductsChanged += () =>
            {
                _logger.Log($"{nameof(OnPurchasedProductsChanged)} has been called.");
                OnPurchasedProductsChanged?.Invoke();
            };

            _subsystems.ForEach<IInitializable>(c => c.Initialize());
        }

        private static PlatformAPI PreparePlatformBridge(EditorDevice editorDevice, string rootPath)
        {
            return
#if !UNITY_EDITOR && UNITY_ANDROID
                new PlatformAndroid(rootPath);
#elif !UNITY_EDITOR && UNITY_IOS
                new PlatformIOS();
#else
                new PlatformDefault(editorDevice);
#endif
        }

        internal void Setup()
        {
            if (_isSetup)
            {
                _logger.Log("Magify client has already been setup.");
                return;
            }
            _isSetup = true;

            Sync();
            _appInstallsTracker.CheckAppInstalls();
        }

        internal void Sync()
        {
            if (!_isSetup)
            {
                _logger.Log("Cannot perform sync before initial setup.");
                return;
            }

            SyncContext();
        }

        private async UniTaskVoid CancelRequestsAndSync()
        {
            await _serverApi.CancelAllServerInteractions();
            Sync();
        }

        private async void SyncContext()
        {
            OnConfigSyncRequested?.Invoke();
            _logger.Log($"{nameof(SyncContext)} has been called");
            if (_serverApi.IsContextRequest)
            {
                _logger.Log($"{nameof(SyncContext)} skipped because context already requested");
                return;
            }

            var context = await _serverApi.GetContextAsync(_disposables.GetOrCreateToken());
            await UniTask.SwitchToMainThread();
            if (context == null)
            {
                _logger.Log($"{nameof(SyncContext)} finished with context: null");
                return;
            }

            _logger.Log($"Load context: ContextBuildTime={context.ContextBuildTime}");
            LastContextSyncTime = new ContextSyncTime(DateTime.UtcNow.ToUnixMilliseconds(), context.ContextBuildTime?.ToUnixMilliseconds() ?? 0);

            SaveContext(context);
            HandleContext(context, ContextKind.Downloaded);

            OnConfigLoaded?.Invoke();
        }

        internal ICampaign CampaignFor(string eventName, Dictionary<string, object> customParams = null, bool silent = false)
        {
            if (silent)
            {
                return _campaignProvider.GetCampaignSilently(eventName, customParams);
            }

            var result = _campaignProvider.ProvideCampaign(eventName, customParams);
            if (result == null)
            {
                return null;
            }

            if (result.IsDefaultCampaign)
            {
                _generalPrefs.UsedDefaultCampaignTypes.Add(result.Campaign.Type.ToEnumString());
            }

            _campaignUpdateHandler.ClearSubscriptions(result.Campaign.Name);
            return result.Campaign;
        }

        internal bool IsCampaignAvailable(string campaignName, string eventName, Dictionary<string, object> customParams = null)
        {
            return _campaignProvider.IsCampaignAvailable(campaignName, eventName, customParams);
        }

        internal IReadOnlyCollection<LtoInfo> GetActiveLtoOffers()
        {
            return _ltoCampaignManager.GetActiveLtoOffers();
        }

        internal void CompleteOffer(string campaignName)
        {
            _ltoCampaignManager.CompleteOffer(campaignName);
        }

        public CampaignImpression GetCampaignImpression(string campaignName)
        {
            return _campaignsTracker.CreateCampaignImpression(campaignName);
        }

        internal List<ContentItem> GetContentList(string group, string key, ICollection<string> tags = null)
        {
            return _contentProvider.GetContentList(group, key, tags);
        }

        internal ContentItem GetEarliestContent(string group, string key)
        {
            return _contentProvider.GetEarliestContent(group, key);
        }

        internal ContentItem GetLatestContent(string group, string key)
        {
            return _contentProvider.GetLatestContent(group, key);
        }

        internal bool HasProcessedPurchase(string productId)
        {
            return _campaignsTracker.AllUsedProducts.Contains(productId);
        }

        internal bool SubscribeCampaignUpdates(ICampaign campaign, Action onUpdate)
        {
            if (_campaignsTracker.GetCampaignRequest(campaign.Name)?.Campaign != campaign) return false;
            _campaignUpdateHandler.SubscribeCampaignUpdates(campaign.Name, onUpdate);
            return true;
        }

        internal void UnsubscribeCampaignUpdates(string campaignName, Action onUpdate)
        {
            _campaignUpdateHandler.UnsubscribeCampaignUpdates(campaignName, onUpdate);
        }

        internal void ClearSubscriptions(string campaignName)
        {
            _campaignUpdateHandler.ClearSubscriptions(campaignName);
        }

        private void ResetCachedConfig()
        {
            _campaignsCollection.ResetCurrentCampaigns();
            _limitsHolder.Reset();
            _ltoCampaignManager.ResetContext();
            _defaultProducts.ResetDefaultProducts();
            _contentProvider.ResetContentMap();
            _featuresProvider.ResetFeatures();

            SetupDefaultContext(_configPath);
        }

        internal void TweakAnalyticsConfig(int eventsGroupSize, int syncTimeInterval)
        {
            _analyticsService.SetupAnalyticsConfig(eventsGroupSize, syncTimeInterval);
        }

        internal void ResetAnalyticsConfig()
        {
            _analyticsService.ResetAnalyticsConfig();
        }

        internal Task<string> ForceReissueAuthToken()
        {
            _serverApi.ResetAuthorizationToken();
            return _serverApi.GetAuthorizationTokenAsync(CancellationToken.None);
        }

        public void RefreshGpsAdId()
        {
            _serverApi.RefreshGpsAdId();
        }

        internal void TweakUserLocale(string languageTag)
        {
            if (_isSetup)
            {
                _contextFileStorage.ResetContext();
                ResetCachedConfig();
                _platform.TweakUserLocale(languageTag);
            }
        }

        internal void SetAttStatus(bool authorized)
        {
            _generalPrefs.AttAuthorizationStatus.Value = authorized;
        }

        internal void SetGameLevel(int level)
        {
            _analyticsService.SetGameLevel(level);
        }

        internal void SetGameMaxLevel(int maxLevel)
        {
            _analyticsService.SetGameMaxLevel(maxLevel);
        }

        internal void SetGameMode(string mode)
        {
            _analyticsService.SetGameMode(mode);
        }

        internal void SetUserEmail(string email)
        {
            _analyticsTracker.TrackMailingStatusMappingEvent(email);
        }

        internal void SetMediaSource(string networkName, string campaignName, string adGroup)
        {
            Debug.Log($"{nameof(SetMediaSource)} with networkName={networkName}, campaignName={campaignName}, adGroup={adGroup}");
            var mediaSource = new MediaSource(networkName, campaignName, adGroup);
            if (_generalPrefs.MediaSource.Equals(mediaSource))
            {
                return;
            }
            _generalPrefs.MediaSource = mediaSource;
            Sync();
        }

        internal void SetupConversionTracker(string revenuePerCountryPath, string revenueLevelsPath, string defaultCurrencyRatesPath, string subscriptionMultipliersPath)
        {
            _logger.Log($"Setup conversion tracker.\n{nameof(revenuePerCountryPath)}: {revenuePerCountryPath}\n{nameof(revenueLevelsPath)}: {revenueLevelsPath}\n{nameof(defaultCurrencyRatesPath)}: {defaultCurrencyRatesPath}\n{nameof(subscriptionMultipliersPath)}: {subscriptionMultipliersPath}");
            _platform.SetupConversionTracker(revenuePerCountryPath, revenueLevelsPath, defaultCurrencyRatesPath, subscriptionMultipliersPath);
        }

        internal void TrackCustomEvent(string eventName, Dictionary<string, object> customParams = null)
        {
            _analyticsService.TrackCustomSessionEvent(eventName, customParams);
        }

        internal void TrackImpression(string campaignName)
        {
            _campaignsTracker.TrackImpression(campaignName);
            _analyticsService.TrackCampaignImpression(campaignName, includeNested: true);
        }

        internal void TrackParentCampaignImpression(string campaignName)
        {
            _campaignsTracker.TrackImpression(campaignName);
            _analyticsService.TrackCampaignImpression(campaignName, includeNested: false);
        }

        internal void TrackProductsImpression(string campaignName, IReadOnlyList<string> productIds)
        {
            _analyticsService.TrackNestedCampaignImpressions(campaignName, productIds);
        }

        internal void TrackImpressionFailFor(string campaignName, string reason)
        {
            _analyticsService.TrackImpressionFailEvent(campaignName, reason);
        }

        internal void TrackAdsImpression(string campaignName, IAdsImpression impression)
        {
            _campaignsTracker.TrackAdsImpression(campaignName);
            _analyticsService.TrackAdsImpression(campaignName, impression);
        }

        internal void TrackClickFor(string campaignName)
        {
            _campaignsTracker.TrackClick(campaignName);
            _analyticsService.TrackClickEvent(EventType.Click, campaignName);
        }

        internal void TrackProductClickFor(string campaignName, string productId)
        {
            _campaignsTracker.TrackClick(campaignName, productId);
            _analyticsService.TrackClickEvent(EventType.Click, campaignName, productId);
        }

        internal void TrackAdsClickFor(string campaignName)
        {
            _campaignsTracker.TrackClick(campaignName);
            _analyticsService.TrackClickEvent(EventType.AdsClick, campaignName);
        }

        internal void TrackAdsProductClickFor(string campaignName, string productId)
        {
            _analyticsService.TrackClickEvent(EventType.AdsClick, campaignName, productId);
        }

        internal void TrackSubscriptionActivation(bool isTrial, string productId, string price, string currency, string transactionId, string purchaseToken, string originalTransactionId, string receipt)
        {
            _analyticsService.TrackSubscription(isTrial, productId, price, currency, transactionId, originalTransactionId);
            _campaignsTracker.TrackSubscriptionPurchase(productId, InAppSourceKind.Internal);
            _featuresProvider.OnProductPurchase();
            _ltoCampaignManager.OnProductPurchase(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
            _purchaseTracker.ValidateSubscription(productId, transactionId, purchaseToken, receipt);
        }

        internal void TrackExternalSubscriptionActivation(bool isTrial, string productId, string price, string currency, string transactionId, string purchaseToken, string originalTransactionId, string receipt)
        {
            _campaignsTracker.TrackSubscriptionPurchase(productId, InAppSourceKind.External);
            _ltoCampaignManager.OnProductPurchase(productId);
            _featuresProvider.OnProductPurchase();

            var purchaseInfo = new PurchaseInfo
            {
                Product = new ProductInfo
                {
                    ProductId = productId,
                    Price = price,
                    Currency = currency
                },
                TransactionId = transactionId,
                OriginalTransactionId = originalTransactionId,
            };

            var eventType = isTrial ? EventType.TrialActivation : EventType.PaidSubscriptionActivation;
            _analyticsTracker.TrackPurchaseEvent(eventType, _campaignsTracker.BuildExternalSubscriptionImpression(), purchaseInfo);
            _purchaseTracker.ValidateSubscription(productId, transactionId, purchaseToken, receipt);
        }

        internal void TrackInAppFor(string productId, string price, string currency, string transactionId, string purchaseToken, string originalTransactionId, string receipt)
        {
            _analyticsService.TrackInApp(isExternal: false, productId, price, currency, transactionId, originalTransactionId);
            _campaignsTracker.TrackInAppPurchase(productId, InAppSourceKind.Internal);
            _featuresProvider.OnProductPurchase();
            _ltoCampaignManager.OnProductPurchase(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
            _purchaseTracker.ValidateInApp(productId, transactionId, purchaseToken, receipt);
        }

        internal void TrackExternalInAppFor(string productId, string price, string currency, string transactionId, string purchaseToken, string originalTransactionId, string receipt)
        {
            _analyticsService.TrackInApp(isExternal: true, productId, price, currency, transactionId, originalTransactionId);
            _campaignsTracker.TrackInAppPurchase(productId, InAppSourceKind.External);
            _featuresProvider.OnProductPurchase();
            _ltoCampaignManager.OnProductPurchase(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
            _purchaseTracker.ValidateInApp(productId, transactionId, purchaseToken, receipt);
        }

        public void TrackRestoredInAppFor(string productId)
        {
            _campaignsTracker.TrackInAppPurchase(productId, InAppSourceKind.External); // restored InApps are considered as external
            _featuresProvider.OnProductPurchase();
            _ltoCampaignManager.OnProductPurchase(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
        }

        public void TrackRestoredSubscription(string productId)
        {
            _campaignsTracker.TrackSubscriptionPurchase(productId, InAppSourceKind.External); // restored Subscriptions are considered as external
            _featuresProvider.OnProductPurchase();
            _ltoCampaignManager.OnProductPurchase(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
        }

        internal void TrackRewardGranted(string productId)
        {
            _campaignsTracker.TrackRewardGranted(productId);
            _ltoCampaignManager.OnRewardGranted(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
        }

        internal void TrackFreeBonusGranted(string productId)
        {
            _campaignsTracker.TrackBonusGranted(productId);
            _ltoCampaignManager.OnBonusGranted(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
        }

        internal void TrackOrdinaryProductUsed(string productId)
        {
            _campaignsTracker.TrackOrdinaryProductUsed(productId);
            _campaignUpdateHandler.OnProductUsed(productId);
        }

        internal void TrackIncomeTransaction(string source, List<BonusInfo> bonuses, ProductInfo product = null)
        {
            _analyticsService.TrackIncomeTransaction(source, bonuses, product);
        }

        internal void TrackExpenseTransaction(List<BonusInfo> bonuses)
        {
            _analyticsService.TrackExpenseTransaction(bonuses);
        }

        internal void TrackCorrectionTransaction(List<BonusInfo> bonuses)
        {
            _analyticsService.TrackCorrectionTransaction(bonuses);
        }

        [CanBeNull]
        internal CampaignRequest GetLastCampaignRequest(string campaignName)
        {
            return _campaignsTracker.GetCampaignRequest(campaignName);
        }

        [CanBeNull]
        internal CampaignRequest GetLastCampaignRequest(CampaignType campaignType)
        {
            return _campaignsTracker.GetCampaignRequest(campaignType);
        }

        private void SetupSavedContext()
        {
            var json = _contextFileStorage.LoadContext();
            var context = CampaignsContext.Deserialize(json);

            if (context == null)
            {
                return;
            }

            HandleContext(context, ContextKind.Saved);
        }

        private void SetupDefaultContext(string configPath)
        {
            var path = Path.Combine(Application.streamingAssetsPath, configPath);
            _logger.Log($"{nameof(SetupDefaultContext)} from file={configPath}");

            string json = default;
            if (_platform.RuntimePlatform == RuntimePlatform.Android)
            {
                var request = UnityWebRequest.Get(path);
                request.SendWebRequest();
                while (!request.isDone)
                {
                }
                json = request.downloadHandler.text;
            }
            else if (File.Exists(path))
            {
                json = File.ReadAllText(path);
            }

            if (string.IsNullOrEmpty(json))
            {
                _logger.LogWarning($"Default config file was not found: {path}");
                return;
            }

            var defaultContext = CampaignsContext.Deserialize(json);
            if (defaultContext == null)
            {
                _logger.LogError($"{nameof(SetupDefaultContext)} from file={configPath} failed. Default context is empty.");
                return;
            }
            _logger.Log($"{nameof(SetupDefaultContext)} from file={configPath} succeeded.");

            _subsystems.ForEach<IContextListener>(c => c.UpdateContext(defaultContext, ContextKind.Default));
        }

        private void SaveContext(CampaignsContext context)
        {
            var json = JsonFacade.SerializeObject(context);
            _contextFileStorage.SaveContext(json);
        }

        public void HandleContext(CampaignsContext context, ContextKind kind)
        {
            _subsystems.ForEach<IContextListener>(c => c.UpdateContext(context, kind));
        }

        private void PrepareForFirstSessionInVersion()
        {
            _logger.Log($"Start of preparing for the current App version: {Application.version}.");
            using (_generalPrefs.MultipleChangeScope())
            {
                _logger.Log("Used Default Campaign Types list will be clear.");
                _generalPrefs.UsedDefaultCampaignTypes.Clear();
                _logger.Log("Authorization Token will be reset.");
                _serverApi.ResetAuthorizationToken();
                _logger.Log("Version counters will be reset.");
                _campaignsTracker.ResetVersionCounters();
            }
            _logger.Log("Preparing for the current App version finished.");
        }

        private static EndpointUrl GetEnvironmentUrl(Environment environment)
        {
            return environment switch
            {
                Magify.Environment.Production => EndpointUrl.Production,
                Magify.Environment.Staging => EndpointUrl.Staging,
                Magify.Environment.Offline => EndpointUrl.Offline,
                _ => throw new InvalidMagifyEnvironmentException(environment)
            };
        }

        public void ResetContext(bool cancelIfLoading)
        {
            if (cancelIfLoading) _serverApi.CancelAllContextLoadings();
            _contextFileStorage.ResetContext();
        }

        public void Reset()
        {
            _generalPrefs.Reset();
            ClearNativeStorage();
        }

        /// <inheritdoc cref="PlatformAPI.ClearNativeStorage"/>
        public void ClearNativeStorage()
        {
            _platform.ClearNativeStorage();
        }

        public void Dispose()
        {
            try
            {
                _disposables.Dispose();
                typeof(MagifyClient).ResetEvents(this);
            }
            catch (Exception e)
            {
                _logger.LogError($"During dispose of MagifyClient an error occurred: {e}");
            }
        }
    }
}