using System.Text;
using System.Threading;
using Magify.Model;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using UnityEngine;
using UnityEngine.Networking;

namespace Magify
{
    internal class NetworkClient : INetworkClient
    {
        [NotNull]
        private static readonly MagifyLogger _logger = MagifyLogger.Get(LoggingScope.Network);

        public async UniTask<WebResponseMessage> SendAsync(string url, WebRequestMessage message, CancellationToken cancellationToken)
        {
            try
            {
                _logger.Log($"SendAsync with message => {message}");
                using var request = CreateRequest(url, message);
                await request.SendWebRequest().WithCancellation(cancellationToken);
                return request.ToWebRequestResponse(message);
            }
            catch (UnityException e)
            {
                _logger.LogError("Failed to send web request with exception." +
                                 "Don't worry, Magify will try to send it bit later again." +
                                 "But you should send info about this error to Magify support." +
                                 $"\n{e}");
                return new WebResponseMessage
                {
                    Result = UnityWebRequest.Result.ConnectionError,
                    ResponseCode = -1,
                    Error = null,
                    Text = null,
                    RequestMessage = message,
                };
            }
        }

        private UnityWebRequest CreateRequest(string url, WebRequestMessage message)
        {
            var payload = new JsonRPCPayload(message.Method, message.Payload);
            var payloadString = JsonFacade.SerializeObject(payload);
            var payloadBytes = new UTF8Encoding().GetBytes(payloadString);

            var request = new UnityWebRequest(url, UnityWebRequest.kHttpVerbPOST)
            {
                uploadHandler = new UploadHandlerRaw(payloadBytes),
                downloadHandler = new DownloadHandlerBuffer(),
                timeout = 35,
            };
            request.SetRequestHeader("Content-Type", "application/json");
            if (message.AuthToken != null)
            {
                request.SetRequestHeader("Authorization", message.AuthToken);
            }

            return request;
        }
    }
}